# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pyfitterbap.pubsub import PubSub


class PubSubTest(unittest.TestCase):

    def setUp(self):
        self.p = PubSub(topic_prefix='hello')
        self.sub1 = []
        self.sub2 = []

    def sub1_fn(self, topic, value, retain=None):
        self.sub1.append((topic, value))

    def sub2_fn(self, topic, value, retain=None):
        self.sub2.append((topic, value))

    def test_sub_pub(self):
        self.p.subscribe('hello/world', self.sub1_fn)
        self.p.publish('hello/world', 'there')
        self.assertEqual([('hello/world', 'there')], self.sub1)

    def test_pub_sub_not_retained(self):
        self.p.publish('hello/world', 'there')
        self.p.subscribe('hello/world', self.sub1_fn)
        self.assertEqual([], self.sub1)

    def test_pub_sub_retained(self):  # retained value
        self.p.publish('hello/world', 'there', retain=True)
        self.p.subscribe('hello/world', self.sub1_fn)
        self.assertEqual([('hello/world', 'there')], self.sub1)

    def test_retained_dedup(self):
        self.p.publish('hello/world', 'there', retain=True)
        self.p.subscribe('hello/world', self.sub1_fn)
        self.p.publish('hello/world', 'there', retain=True)
        self.assertEqual([('hello/world', 'there')], self.sub1)

    def test_pub_src(self):
        self.p.subscribe('hello/world', self.sub1_fn)
        self.p.publish('hello/world', 'there', src_cbk=self.sub1_fn)
        self.assertEqual([], self.sub1)

    def test_sub_parent(self):
        self.p.subscribe('', self.sub1_fn)
        self.p.subscribe('hello/there', self.sub2_fn)
        self.p.publish('hello/there/world', 'value')
        self.assertEqual([('hello/there/world', 'value')], self.sub1)
        self.assertEqual([('hello/there/world', 'value')], self.sub1)

    def test_unsub(self):
        self.p.subscribe('hello/world', self.sub1_fn)
        self.p.unsubscribe('hello/world', self.sub1_fn)
        self.p.publish('hello/world', 'there')
        self.assertEqual([], self.sub1)

    def test_get(self):
        self.p.publish('hello/world', 'there', retain=True)
        self.assertEqual('there', self.p.get('hello/world'))
        self.p.publish('hello/world', 'new', retain=True)
        self.assertEqual('new', self.p.get('hello/world'))
        self.p.publish('hello/world', 'newer')  # not retained!
        self.assertIsNone(self.p.get('hello/world'))

    def test_meta(self):
        meta1 = {'type': 'u32'}
        self.p.meta('hello/world', meta1)
        self.p.subscribe('', self.sub1_fn, forward=True)
        self.p.publish('$', None)
        self.assertEqual([('hello/world$', meta1), ('$', None)], self.sub1)

    def test_meta_other_subscribe_root(self):
        self.p.subscribe('', self.sub1_fn, forward=True)
        self.p.publish('other/$', None)
        self.assertEqual([('other/$', None)], self.sub1)

    def test_meta_other_subscribe_subtopic(self):
        self.p.subscribe('other', self.sub1_fn, forward=True)
        self.p.publish('other/$', None)
        self.assertEqual([('other/$', None)], self.sub1)

    # def test_query(self):  # todo




from .version import __version__, __title__, __description__, __url__, \
    __author__, __author_email__, __license__, __copyright__


#from . import lib
#from . import collections, ec, memory, stream
#from .pattern_32a import allocate_pattern_buffer, PatternTx, PatternRx



# Copyright 2020-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import serial.tools.list_ports
from PySide6 import QtCore, QtGui, QtWidgets


BAUDRATES = [9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600, 1000000, 2000000, 3000000, 5000000]
BAUDRATE_DEFAULT_IDX = 10


class OpenDialog(QtWidgets.QDialog):

    def __init__(self, parent=None):
        super(OpenDialog, self).__init__(parent)
        self.setObjectName('OpenDialog')
        self.setWindowTitle('Open')
        self.resize(259, 140)
        self._main_vertical_layout = QtWidgets.QVBoxLayout(self)
        self._main_vertical_layout.setObjectName('open_main_layout')

        self._layout = QtWidgets.QGridLayout()
        self._layout.setObjectName('open_inner_layout')

        self._device_label = QtWidgets.QLabel(self)
        self._device_label.setObjectName('open_device_label')
        self._device_label.setText("Device")
        self._layout.addWidget(self._device_label, 0, 0, 1, 1)

        self._device_combo_box = QtWidgets.QComboBox(self)
        self._device_combo_box.setObjectName('open_device_combo_box')
        for device in self._devices():
            self._device_combo_box.addItem(device)
        self._layout.addWidget(self._device_combo_box, 0, 1, 1, 1)

        self._baud_label = QtWidgets.QLabel(self)
        self._baud_label.setObjectName('open_baud_label')
        self._baud_label.setText('Baud')
        self._layout.addWidget(self._baud_label, 1, 0, 1, 1)

        self._baud_combo_box = QtWidgets.QComboBox(self)
        self._baud_combo_box.setObjectName("open_device_combo_box")
        for baud in BAUDRATES:
            self._baud_combo_box.addItem(str(baud))
        self._baud_combo_box.setEditable(True)
        self._baud_combo_box.setCurrentIndex(BAUDRATE_DEFAULT_IDX)
        self._baud_validator = QtGui.QIntValidator(0, 10000000, self)
        self._baud_combo_box.setValidator(self._baud_validator)
        self._layout.addWidget(self._baud_combo_box, 1, 1, 1, 1)

        self._main_vertical_layout.addLayout(self._layout)

        self.buttonBox = QtWidgets.QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel | QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self._main_vertical_layout.addWidget(self.buttonBox)

        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), self.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), self.reject)

    def _devices(self):
        comports = serial.tools.list_ports.comports()
        devices = [c.device for c in comports]
        if len(devices) and devices[0].startswith('COM'):
            ports = sorted([int(d[3:]) for d in devices])
            devices = [f'COM{port}' for port in ports]
        else:
            devices = sorted(devices)
        return devices

    def exec_(self):
        if QtWidgets.QDialog.exec_(self) == 1:
            return {
                'device': self._device_combo_box.currentText(),
                'baud': int(self._baud_combo_box.currentText()),
            }
        else:
            return None


# Copyright 2020-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from pyfitterbap import crc


def parser_config(p):
    """compute CRC."""
    p.add_argument('--data',
                   help='The CRC data.')
    return on_cmd


def on_cmd(args):
    if args.data is not None:
        x = np.array([int(x, 0) for x in args.data.split(',')], dtype=np.uint8)
        y = crc.crc32(0, x)
        print(f'0x{y:08x}')
        return 0
    return 1




# Copyright 2020-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from PySide6 import QtCore, QtGui, QtWidgets
from queue import Queue, Empty


log = logging.getLogger(__name__)


class _ResyncEvent(QtCore.QEvent):
    """An event containing a request for a resync function call."""
    EVENT_TYPE = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

    def __init__(self):
        QtCore.QEvent.__init__(self, self.EVENT_TYPE)

    def __str__(self):
        return '_ResyncEvent()'

    def __len__(self):
        return 0


class Resync(QtCore.QObject):
    """Resynchronize to the Qt event thread.

    :param parent: The parent QObject.

    Qt does not easily support resynchronization from non-Qt threads.  See
    this [bug report](https://github.com/jetperch/pyjoulescope_ui/issues/69).
    This class implements a resynchronization service that allows a non-Qt
    thread to invoke a method on the main Qt event thread.

    See [this page](https://doc.qt.io/qt-5/threads-qobject.html) for a
    Qt threading overview.
    """
    def __init__(self, parent=None):
        super(Resync, self).__init__(parent)
        self._resync_handlers = {}
        self._resync_queue = Queue()

    def event(self, event: QtCore.QEvent):
        if event.type() == _ResyncEvent.EVENT_TYPE:
            # process our resync calls.
            event.accept()
            try:
                fn, args, kwargs, ev = self._resync_queue.get(timeout=0.0)
                if id(event) != id(ev):
                    log.warning('event mismatch')
                fn(*args, **kwargs)
            except Empty:
                log.warning('event signaled but not available')
            except Exception:
                log.exception('resync queue failed')
            return True
        else:
            return super(Resync, self).event(event)

    def _resync_handle(self, name, args, kwargs):
        # safely resynchronize to the main Qt event thread
        event = _ResyncEvent()
        self._resync_queue.put((name, args, kwargs, event))
        QtCore.QCoreApplication.postEvent(self, event)

    def clear(self, target=None):
        """Clear handlers.

        :param target: The target provided to :meth:`factory`.  If None,
            then clear all targets.
        """
        if target is not None:
            key = id(target)
            self._resync_handlers.pop(key)
        else:
            self._resync_handlers.clear()

    def wrap(self, target):
        """Get a function that will call target from the main Qt event loop.

        :param target: The target callable(args, kwargs).  This instance will
            hold a reference to target.
        :return: The resynchronization function that invokes target from the
            main thread's Qt event loop.  Calls to this function
            complete immediately, but the actual processing is deferred.
        """
        def fn(*args, **kwargs):
            return self._resync_handle(target, args, kwargs)
        key = id(target)
        if key not in self._resync_handlers:
            self._resync_handlers[key] = fn
        return self._resync_handlers[key]



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from PySide6 import QtCore, QtGui, QtWidgets


class ExpandingWidget(QtWidgets.QWidget):

    def __init__(self, parent=None, title=None, animation_duration_ms=None):
        super(ExpandingWidget, self).__init__(parent=parent)
        self._widget: QtWidgets.QWidget = None
        self._animation_duration_ms = 250 if animation_duration_ms is None else int(animation_duration_ms)
        self._animation = None
        self._main_layout = QtWidgets.QVBoxLayout(self)
        self._main_layout.setContentsMargins(0, 0, 0, 0)

        self._toggle_button = QtWidgets.QToolButton()
        self._toggle_button.setStyleSheet("QToolButton { border: none; }")
        self._toggle_button.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self._toggle_button.setArrowType(QtCore.Qt.RightArrow)
        self._toggle_button.setCheckable(True)
        self._toggle_button.setChecked(False)
        self.title = title
        self._main_layout.addWidget(self._toggle_button)

        self._toggle_button.clicked.connect(self._start_animation)

    @property
    def title(self):
        return self._toggle_button.text()

    @title.setter
    def title(self, title):
        title = '' if title is None else str(title)
        self._toggle_button.setText(title)

    def _start_animation(self, checked):
        if self._widget is None:
            return
        if checked:
            self._widget.show()
        else:
            self._widget.hide()
        arrow_type = QtCore.Qt.DownArrow if checked else QtCore.Qt.RightArrow
        self._toggle_button.setArrowType(arrow_type)

    def _show(self):
        if self._widget is None:
            return
        # See https://www.qtcentre.org/threads/60494-Animate-an-hidden-widget
        if self._widget.isHidden():
            self._widget.show()   # required to get size
            pos = self._widget.pos()
            size = self._widget.size()
            fo = QtCore.QRect(pos.x(), pos.y() + size.height(), size.width(), size.height())
            fi = QtCore.QRect(pos, size)
        else:
            fo = self._widget.geometry()
            fi = QtCore.QRect(fo.x(), fo.y() - fo.height(), fo.width(), fo.height())

        animation = QtCore.QPropertyAnimation(self._widget, b'geometry')
        animation.setDuration(self._animation_duration_ms)
        animation.setEasingCurve(QtCore.QEasingCurve.Linear)
        animation.setStartValue(fo)
        animation.setEndValue(fi)
        animation.start()
        self._animation = animation

    def setWidget(self, widget):
        self._widget = widget
        widget.setParent(self)
        self._main_layout.addWidget(self._widget)
        self._widget.setHidden(not self._toggle_button.isChecked())


class VerticalScrollArea(QtWidgets.QScrollArea):

    def __init__(self, parent=None):
        super(VerticalScrollArea, self).__init__(parent)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(QtGui.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtGui.Qt.ScrollBarAsNeeded)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                           QtWidgets.QSizePolicy.Preferred)

    def setWidget(self, widget):
        widget.installEventFilter(self)
        return super(VerticalScrollArea, self).setWidget(widget)

    def eventFilter(self, obj, event):
        widget = self.widget()
        if obj == widget and event.type() == QtCore.QEvent.Resize:
            width = widget.minimumSizeHint().width() + self.verticalScrollBar().width()
            self.setMinimumWidth(width)
        return False


class _MainWindow(QtWidgets.QMainWindow):

    def __init__(self):
        super(_MainWindow, self).__init__()
        self.setWindowTitle('ExpandingWidget Demo')
        self._scroll_widget = VerticalScrollArea(self)
        self._scroll_widget.setObjectName('central_scroll')
        self.setCentralWidget(self._scroll_widget)

        self._widget = QtWidgets.QWidget(self._scroll_widget)
        self._widget.setObjectName('central_widget')
        self._scroll_widget.setWidget(self._widget)

        self._layout = QtWidgets.QVBoxLayout(self._widget)
        self._layout.setSpacing(6)
        self._layout.setContentsMargins(11, 11, 11, 11)
        self._layout.setObjectName('central_layout')

        self._widgets = []
        for widget_id in range(5):
            widget = ExpandingWidget(self._widget, title=f'Widget {widget_id}')
            inner_widget = QtWidgets.QWidget(widget)
            layout = QtWidgets.QVBoxLayout(inner_widget)
            labels = []
            for label_id in range(10):
                label = QtWidgets.QLabel(f'Widget {widget_id}, Label {label_id}', inner_widget)
                layout.addWidget(label)
                labels.append(label)
            widget.setWidget(inner_widget)
            self._widgets.append([widget, inner_widget, layout, labels])
            self._layout.addWidget(widget)

        self._spacer = QtWidgets.QSpacerItem(0, 0,
                                             QtWidgets.QSizePolicy.Minimum,
                                             QtWidgets.QSizePolicy.Expanding)
        self._layout.addItem(self._spacer)
        self.show()


def _run():
    import sys
    import ctypes
    # http://doc.qt.io/qt-5/highdpi.html
    # https://vicrucann.github.io/tutorials/osg-qt-high-dpi/
    if sys.platform.startswith('win'):
        ctypes.windll.user32.SetProcessDPIAware()
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    app = QtWidgets.QApplication(sys.argv)
    ui = _MainWindow()
    rc = app.exec_()
    del ui
    del app
    return rc


if __name__ == '__main__':
    _run()



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pyqtgraph as pg
from PySide6 import QtCore, QtGui, QtWidgets
import logging
import numpy as np


class WaveformWidget(QtWidgets.QWidget):

    def __init__(self, parent=None, pubsub=None):
        super(WaveformWidget, self).__init__(parent)
        self._topics = {}
        self._log = logging.getLogger(__name__)

        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._source_widget = QtWidgets.QWidget(self)
        self._source_layout = QtWidgets.QHBoxLayout(self._source_widget)
        self._source_label = QtWidgets.QLabel('Source: ', parent=self._source_widget)
        self._source_layout.addWidget(self._source_label)
        self._source_combobox = QtWidgets.QComboBox(self._source_widget)
        self._source_layout.addWidget(self._source_combobox)
        self._layout.addWidget(self._source_widget)

        self.win = pg.GraphicsLayoutWidget(parent=self, show=True, title="Waveform")
        self.win.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        # self.win.sceneObj.sigMouseClicked.connect(self._on_mouse_clicked_event)
        # self.win.sceneObj.sigMouseMoved.connect(self._on_mouse_moved_event)
        self._layout.addWidget(self.win)

        self._plot: pg.PlotItem = self.win.addPlot(row=0, col=0)
        self._plot.showGrid(True, True, 128)
        self._curve = self._plot.plot(pen='y')

        self._data = np.empty(32000, dtype=np.float32)
        self._data[:] = np.nan
        self._buffer = np.empty(32000, dtype=np.float32)
        self._index = 0
        self._data_pending = False

        if pubsub is not None:
            pubsub.subscribe('', self._on_update, skip_retained=True)

        self._timer = QtCore.QTimer()
        self._timer.timeout.connect(self._redraw)
        self._timer.start(50)

    def _redraw(self):
        if self._data_pending:
            self._data_pending = False
            self._curve.setData(self._data)

    def _on_update(self, topic, value, retain):
        if topic.endswith('/din'):
            v = self._topics.get(topic, 0)
            if v == 0:
                self._log.info('Added waveform source: %s', topic)
                self._source_combobox.addItem(topic)
            self._topics[topic] = v + 1
            if topic == self._source_combobox.currentText():
                sample_id_bytes = value[:4]
                d = self._buffer
                d_next = np.frombuffer(value[4:], dtype=np.float32)
                d_len = len(d_next)
                d_end = self._index + d_len
                if d_end > len(d):
                    d_len = len(d) - self._index
                    d[self._index:] = d_next[:d_len]
                    self._data, self._buffer = self._buffer, self._data
                    self._data_pending = True
                    self._index = 0
                else:
                    d[self._index:(self._index + d_len)] = d_next
                    self._index += d_len



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from PySide6 import QtCore, QtGui, QtWidgets
import logging


class VerticalScrollArea(QtWidgets.QScrollArea):

    def __init__(self, parent=None):
        super(VerticalScrollArea, self).__init__(parent)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(QtGui.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtGui.Qt.ScrollBarAsNeeded)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                           QtWidgets.QSizePolicy.Preferred)
        self._log = logging.getLogger(__name__)

    def setWidget(self, widget):
        widget.installEventFilter(self)
        return super(VerticalScrollArea, self).setWidget(widget)

    def eventFilter(self, obj, event):
        widget = self.widget()
        if obj == widget and event.type() == QtCore.QEvent.Resize:
            width = widget.minimumSizeHint().width() + self.verticalScrollBar().width()
            self.setMinimumWidth(width)
            self._log.debug('scroll: sizeHint=%s, minimumSizeHint=%s, minimumSize=%s',
                            self.sizeHint(), self.minimumSizeHint(), self.minimumSize())
        return False



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import weakref
from collections.abc import Mapping


def _as_bool(x):
    if isinstance(x, str):
        x = x.lower()
    if x in [0, 'no', 'off', 'disable', 'disabled', 'false', 'inactive']:
        return False
    if x in [1, 'yes', 'on', 'enable', 'enabled', 'true', 'active']:
        return True
    raise ValueError(f'Invalid boolean value: {x}')


class _Topic:

    def __init__(self, parent, topic, value=None):
        """Hold a single Topic entry for :class:`PubSub`.

        :param parent: The parent :class:`Topic` instance.
        :param topic: The topic string.
        :param value: The optional initial value for the topic.
        """
        self._topic = topic
        self._value = value
        self.meta = None
        if parent is not None:
            parent = weakref.ref(parent)
        self.parent = parent
        self.children: Mapping[str, _Topic] = {}
        self._subscribers = []  # list of tuples of callback, forward

    def __str__(self):
        return f'Topic({self._topic}, value={self._value})'

    @property
    def name(self):
        return self._topic

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, x):
        self.publish(x)

    def publish(self, x, retain=None, src_cbk=None):
        if bool(retain):
            if self._value == x:  # deduplicate
                return
            self._value = x
        else:
            self._value = None
        topic = self
        while topic is not None:
            for subscriber, forward in topic._subscribers:
                if subscriber == src_cbk:
                    continue
                subscriber(self._topic, x, retain)
            topic = topic.parent
            if topic is not None:
                topic = topic()  # weakref to parent

    def forward(self, topic, value, retain=None, src_cbk=None, traverse_parent=False):
        for subscriber, forward in self._subscribers:
            if subscriber == src_cbk or not forward:
                continue
            subscriber(topic, value, retain)
        if traverse_parent and self.parent is not None:
            parent = self.parent()
            if parent is not None:
                parent.forward(topic, value, retain, src_cbk, traverse_parent)

    def _publish_new_subscriber(self, cbk):
        for child in self.children.values():
            child._publish_new_subscriber(cbk)
        if self._value is not None:
            cbk(self._topic, self._value)

    def subscribe(self, cbk, skip_retained=None, forward=None):
        if not callable(cbk):
            raise ValueError('subscribers must be callable')
        if cbk not in self._subscribers:
            self._subscribers.append((cbk, forward))
        if not bool(skip_retained):
            self._publish_new_subscriber(cbk)

    def unsubscribe(self, cbk):
        unsub = []
        for idx, (cbk_fn, _) in enumerate(self._subscribers):
            if cbk == cbk_fn:
                unsub.append(idx)
        for idx in reversed(unsub):
            self._subscribers.pop(idx)


class PubSub:
    """A trivial, local publish/subscribe implementation.

    :param topic_prefix: The topic prefix string owned by this
        implementation, which is used for distributed pubsub.
        Provide '' or None if you are using a single instance.

    This PubSub implementation features:
    * retained values
    * de-duplication for retained values
    * Support for cascaded distributed instances
    * Omit publisher on publish

    Topic names are hierarchical using '/' separators.
    Request topic metadata using '$' or 'my/path/$'.
    Metadata is returned as 'my/path/var$'
    Query the current value using '?' or 'my/path/?'.
    The query values are returned as 'my/path/var?'
    """
    def __init__(self, topic_prefix: str = None):
        self._topic_prefix = '' if topic_prefix is None else str(topic_prefix)
        self._root = _Topic(None, '')

    def _topic_find(self, topic, create=False):
        t = self._root
        while len(topic) and topic[-1] in '/$?':
            topic = topic[:-1]
        if len(topic):
            parts_so_far = []
            parts = topic.split('/')
            for part in parts:
                parts_so_far.append(part)
                child = t.children.get(part)
                if child is None:
                    if create:
                        subtopic = '/'.join(parts_so_far)
                        child = _Topic(t, subtopic)
                        t.children[part] = child
                    else:
                        return None
                t = child
        return t

    def _topic_find_existing_base(self, topic):
        t = self._root
        while len(topic) and topic[-1] in '/$?':
            topic = topic[:-1]
        if len(topic):
            parts_so_far = []
            parts = topic.split('/')
            for part in parts:
                parts_so_far.append(part)
                child = t.children.get(part)
                if child is None:
                    return t
                t = child
        return t

    def _meta_send_all(self, topic: _Topic, src_cbk=None):
        if topic is None:
            return
        if topic.meta is not None:
            topic.forward(topic.name + '$', topic.meta, retain=False, src_cbk=src_cbk, traverse_parent=True)
        for child in topic.children.values():
            self._meta_send_all(child, src_cbk)

    def publish(self, topic, value, retain=None, src_cbk=None):
        """Publish to a topic.

        :param topic: The topic name.
        :param value: The value for the topic.
        :param retain: True to retain (hold on to) value, False publish & discard.
            None (default) is False.
        :param src_cbk: The subscriber that will not be updated.
            None (default) updates all applicable subscribers.
        """
        if not len(topic):
            raise ValueError('Empty topic not allowed')
        if topic == '$':
            self._meta_send_all(self._topic_find(self._topic_prefix, create=True), src_cbk)
            self._root.forward(topic, None, retain=retain, src_cbk=src_cbk)
        elif topic.endswith('/$'):
            if topic.startswith(self._topic_prefix):  # for us
                self._meta_send_all(self._topic_find(topic, create=True), src_cbk)
            else:
                t = self._topic_find_existing_base(topic)
                t.forward(topic, None, retain=retain, src_cbk=src_cbk, traverse_parent=True)
        elif topic[-1] == '$':  # endswith
            t = self._topic_find(topic[:-1], create=True)
            if topic.startswith(self._topic_prefix):  # for use
                if value is None:  # request for us
                    if t.value is not None:
                        t.forward(topic, t.value, retain=True, src_cbk=src_cbk, traverse_parent=True)
                else:  # set for us
                    t.meta = value
                    t.forward(topic, value, retain=True, src_cbk=src_cbk, traverse_parent=True)
            else:  # get or set for another pubsub instance relay
                t.forward(topic, value, retain=True, src_cbk=src_cbk, traverse_parent=True)
        elif topic[-1] == '?':
            pass  # todo, topic query
        else:
            t = self._topic_find(topic, create=True)
            t.publish(value, retain, src_cbk)

    def meta(self, topic, meta):
        if topic[-1] != '$':
            topic = topic + '$'
        self.publish(topic, meta)

    def get(self, topic):
        """Get the retained value for the topic.

        :param topic: The topic name.
        :return: The topic retained value.
        :raise KeyError: If topic does not exist.
        """
        t = self._topic_find(topic, create=False)
        if t is None:
            raise KeyError(f'topic {topic} does not exist')
        return t.value

    def subscribe(self, topic, cbk, skip_retained=None, forward=None):
        """Subscribe to a topic and its children.

        :param topic: The topic name.
        :param cbk: The callable(topic, value, retain) called on value changes.
        :param skip_retained: Skip the update of all retained values
            that normally occurs when subscribing.
        :param forward: When true, forward $ and ? requests to this subscriber.
            Use True for distributed PubSub instances and user interfaces.
        """
        t = self._topic_find(topic, create=True)
        t.subscribe(cbk, skip_retained=skip_retained, forward=forward)

    def unsubscribe(self, topic, cbk):
        """Unsubscribe from a topic.

        :param topic: The topic name.
        :param cbk: The callable provided to subscribe.
        """
        t = self._topic_find(topic, create=True)
        t.unsubscribe(cbk)

    def create(self, topic, meta=None, subscriber_cbk=None, skip_retained=None, forward=None):
        t = self._topic_find(topic, create=False)
        if t is not None:
            raise ValueError(f'Topic {topic} already exists')
        t = self._topic_find(topic, create=True)
        t.meta = meta
        if meta is not None and 'default' in meta:
            retain = _as_bool(meta.get('retain', 0))
            x = meta['default']
            t.publish(x, retain, subscriber_cbk)
        if subscriber_cbk is not None:
            t.subscribe(subscriber_cbk, skip_retained=skip_retained, forward=forward)



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dateutil.parser
import datetime


FBP_TIME_EPOCH_UNIX_OFFSET_SECONDS = dateutil.parser.parse('2018-01-01T00:00:00Z').timestamp()
Q_GAIN = 2 ** 30


def now():
    k = datetime.datetime.utcnow().timestamp()
    rel_s = k - FBP_TIME_EPOCH_UNIX_OFFSET_SECONDS
    return int(rel_s * Q_GAIN)



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from PySide6 import QtCore, QtGui, QtWidgets
from .expanding_widget import ExpandingWidget
from pyfitterbap.comm.comm import Port0Events
import logging
import weakref

PORTS_COUNT = 32  # todo
DTYPES = ['str', 'json', 'bin',
          'f32', 'f64',
          'bool',
          'u8', 'u16', 'u32', 'u64',
          'i8', 'i16', 'i32', 'i64']
FLAGA = ['ro', 'hide', 'dev']
log = logging.getLogger(__name__)
SPACING = 5
MARGINS = (15, 0, 5, 5)

V_MIN = {
    'u8': 0,
    'u16': 0,
    'u32': 0,
    'u64': 0,
    'i8': -2**7,
    'i16': -2**15,
    'i32': -2**31,
    'i64': -2**64,
}


V_MAX = {
    'u8': 2**8 - 1,
    'u16': 2**16 - 1,
    'u32': 2**32 - 1,
    'u64': 2**64 - 1,
    'i8': 2**7 - 1,
    'i16': 2**15 - 1,
    'i32': 2**31 - 1,
    'i64': 2**64 - 1,
}


class DeviceWidget(QtWidgets.QWidget):

    def __init__(self, parent, device):
        super(DeviceWidget, self).__init__(parent)
        self._device = device
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setSpacing(SPACING)
        self._layout.setContentsMargins(5, 0, 5, 5)
        self._layout.setObjectName('device_layout')

        self._port_expander = ExpandingWidget(self, 'Ports')
        self._port_widget = PortWidget(self._port_expander)
        self._port_expander.setWidget(self._port_widget)
        self._layout.addWidget(self._port_expander)
        device.subscribe('h/c/port', self._port_widget.on_port_meta)

        self._status_expander = ExpandingWidget(self, 'Status')
        self._status_widget = StatusWidget(self._status_expander)
        self._status_expander.setWidget(self._status_widget)
        self._layout.addWidget(self._status_expander)

        # self._echo_widget = EchoWidget(self, device)
        # self._layout.addWidget(self._echo_widget)

        self._pubsub_expander = ExpandingWidget(self, 'PubSub')
        self._pubsub_widget = PubSubWidget(self._pubsub_expander, device)
        self._pubsub_expander.setWidget(self._pubsub_widget)
        self._layout.addWidget(self._pubsub_expander)

        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                           QtWidgets.QSizePolicy.Preferred)
        # print(f'device sizeHint={self.sizeHint()}, minimumSize={self.minimumSize()}')

    def minimumSize(self):
        return self.sizeHint()

    def status_update(self, status):
        self._status_widget.update(status)


class PortWidget(QtWidgets.QWidget):

    def __init__(self, parent=None):
        self._outstanding = 8
        self._tx_port_id = 0
        self._rx_port_id = 0
        self._send_fn = None
        super(PortWidget, self).__init__(parent)
        self.setObjectName('port_widget')
        self.setGeometry(QtCore.QRect(0, 0, 294, 401))
        self.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)

        self._layout = QtWidgets.QGridLayout(self)
        self._layout.setSpacing(SPACING)
        self._layout.setContentsMargins(*MARGINS)
        self._layout.setObjectName('port_widget_layout')

        port_id_hdr = QtWidgets.QLabel('Port ID', self)
        type_hdr = QtWidgets.QLabel('Type', self)
        name_hdr = QtWidgets.QLabel('Name', self)
        self._hdr = self._add_row(0, port_id_hdr, type_hdr, name_hdr)

        self._items = []
        for idx in range(PORTS_COUNT):
            port_id_label = QtWidgets.QLabel(self)
            port_id_label.setObjectName(f'port_id_label_{idx}')
            port_id_label.setText(f'{idx}')
            type_label = QtWidgets.QLabel(self)
            type_label.setObjectName(f'port_type_label_{idx}')
            type_label.setText(' ')
            name_label = QtWidgets.QLabel(self)
            name_label.setObjectName(f'port_name_label_{idx}')
            name_label.setText(' ')
            self._items.append(self._add_row(idx + 1, port_id_label, type_label, name_label))

    def _add_row(self, row, port_id_label, type_label, name_label):
        labels = port_id_label, type_label, name_label
        for idx, label in enumerate(labels):
            label.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
            self._layout.addWidget(label, row, idx, 1, 1)
        return labels

    def on_port_meta(self, topic, value, **kwargs):
        if not topic.endswith('/meta'):
            return
        topic_parts = topic.split('/')
        port_id = int(topic_parts[-2])
        type_label, name_label = self._items[port_id][1:]
        if value is None:
            txt = '-'
            name = ''
        else:
            txt = value.get('type', 'unknown')
            name = value.get('name', '')
        log.info("%d, %s : %s => %s", port_id, topic, value, txt)
        type_label.setText(txt)
        name_label.setText(name)

    def eventFilter(self, obj, event):
        if obj == self:
            print(event)


class StatusWidget(QtWidgets.QWidget):

    def __init__(self, parent):
        super(StatusWidget, self).__init__(parent)
        self.setObjectName('status')
        self._layout = QtWidgets.QGridLayout(self)
        self._layout.setSpacing(SPACING)
        self._layout.setContentsMargins(*MARGINS)
        self._layout.setObjectName('status_layout')
        self._prev = None
        self._items = {}

    def clear(self):
        while not self._layout.isEmpty():
            item = self._layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.setParent(None)
        self._items.clear()
        self._prev = None

    def _statistic_get(self, name):
        try:
            return self._items[name]
        except KeyError:
            row = self._layout.rowCount()
            n = QtWidgets.QLabel(self)
            n.setText(name)
            v = QtWidgets.QLabel(self)
            self._layout.addWidget(n, row, 0)
            self._layout.addWidget(v, row, 1)
            value = (n, v)
            self._items[name] = value
            return value

    def _statistic_update(self, name, value):
        _, v = self._statistic_get(name)
        v.setText(str(value))

    def update(self, status):
        for top_key, top_obj in status.items():
            if isinstance(top_obj, dict):
                for key, obj in top_obj.items():
                    self._statistic_update(f'{top_key}.{key}', obj)
            else:
                self._statistic_update(top_key, top_obj)
        if self._prev is not None:
            rx_bytes = status['rx']['msg_bytes'] - self._prev['rx']['msg_bytes']
            tx_bytes = status['tx']['msg_bytes'] - self._prev['tx']['msg_bytes']
            self._statistic_update('Δrx.msg_bytes', rx_bytes)
            self._statistic_update('Δtx.msg_bytes', tx_bytes)
        self._prev = status


class EchoWidget(QtWidgets.QWidget):

    def __init__(self, parent, device):
        self._device = weakref.ref(device)
        super(EchoWidget, self).__init__(parent)

        self.setObjectName('echo')
        self.setGeometry(QtCore.QRect(0, 0, 294, 401))

        self._layout = QtWidgets.QGridLayout(self)
        self._layout.setSpacing(SPACING)
        self._layout.setContentsMargins(*MARGINS)
        self._layout.setObjectName('echo_layout')

        self._outstanding_label = QtWidgets.QLabel(self)
        self._outstanding_label.setObjectName('echo_outstanding_label')
        self._outstanding_label.setText('Outstanding frames')
        self._layout.addWidget(self._outstanding_label, 0, 0, 1, 1)

        self._outstanding_combo_box = QtWidgets.QComboBox(self)
        self._outstanding_combo_box.setObjectName('echo_outstanding_combobox')
        for frame in [1, 2, 4, 8, 16, 32]:
            self._outstanding_combo_box.addItem(str(frame))
        self._outstanding_combo_box.setEditable(True)
        self._outstanding_combo_box.setCurrentIndex(3)
        self._outstanding_validator = QtGui.QIntValidator(0, 256, self)
        self._outstanding_combo_box.setValidator(self._outstanding_validator)
        self._layout.addWidget(self._outstanding_combo_box, 0, 1, 1, 1)

        self._button = QtWidgets.QPushButton(self)
        self._button.setCheckable(True)
        self._button.setText('Press to start')
        self._button.toggled.connect(self._on_button_toggled)
        self._layout.addWidget(self._button, 1, 0, 2, 0)

    def _on_echo_enabled(self, topic, value):
        self._button.setChecked(value)

    def _on_echo_outstanding(self, topic, value):
        self._outstanding_combo_box.setCurrentText(str(value))

    def _on_button_toggled(self, checked):
        log.info('echo button  %s', checked)
        txt = 'Press to stop' if checked else 'Press to start'
        self._button.setText(txt)
        device = self._device()
        if device is not None:
            value = 1 if checked else 0
            device.publish('h/c/port/0/echo/enable', value, retain=True, src_cbk=self._on_echo_enabled)


class Value(QtCore.QObject):

    def __init__(self, parent, topic, meta):
        QtCore.QObject.__init__(self, parent)
        self._parent = parent
        self._topic = topic
        self._meta = meta
        self._value = None
        self._options = None
        self.label = None
        self.editor = None
        # See include/fitterbap/pubsub.md for the latest format information
        dtype = meta.get('dtype')
        if dtype not in DTYPES:
            log.warning('topic %s: unsupported dtype %s', topic, dtype)
            return
        name = topic
        brief = meta.get('brief')
        detail = meta.get('detail')
        default = meta.get('default')
        options = meta.get('options')
        flags = meta.get('flags')
        flags = [] if flags is None else flags
        # print(f'META: topic={topic}, dtype={dtype}, brief={brief}, detail={detail}, default={default}, options={options}, flags={flags}')

        self.label = QtWidgets.QLabel(name, parent)
        tooltip = f'<html><body><p>{brief}</p>'
        if detail is not None:
            tooltip += f'<p>{detail}</p>'
        tooltip += f'</body></html>'
        self.label.setToolTip(tooltip)

        if 'ro' in flags:
            self.editor = QtWidgets.QLabel(parent)
        elif options is not None and len(options):
            self._options = []
            self.editor = QtWidgets.QComboBox(parent)
            for idx, opts in enumerate(options):
                v = opts[0]
                if len(opts) > 1:
                    e = str(opts[1])
                else:
                    e = str(v)
                self._options.append((v, e))
                self.editor.addItem(e)
            self.editor.currentIndexChanged.connect(self._on_combobox)
        elif dtype == 'str':
            self.editor = QtWidgets.QTextEdit(parent)
        elif dtype == 'json':
            self.editor = QtWidgets.QLabel("json: unsupported", parent)
        elif dtype == 'bin':
            self.editor = QtWidgets.QLabel("bin: unsupported", parent)
        #elif dtype in ['f32', 'f64']:
        #    pass
        elif dtype in ['u8', 'u16', 'u32', 'u64', 'i8', 'i16', 'i32', 'i64']:
            v_min, v_max, v_step = V_MIN[dtype], V_MAX[dtype], 1
            drange = self._meta.get('range')
            if drange is None:
                pass
            elif len(drange) == 2:
                v_min, v_max = drange
            elif len(drange) == 3:
                v_min, v_max, v_step = drange
            else:
                raise RuntimeError('topic %s: invalid range %s', self._topic, drange)
            self.editor = QtWidgets.QSpinBox(parent)
            if 'ro' not in flags:
                self.editor.setRange(v_min, v_max)
                self.editor.setSingleStep(v_step)
            self.editor.valueChanged.connect(self._on_spinbox)
        elif dtype == 'bool':
            self.editor = QtWidgets.QCheckBox(parent)
            self.editor.clicked.connect(self._on_clicked)
        else:
            self.editor = QtWidgets.QLabel("value", parent)
        self.editor.setToolTip(tooltip)
        if 'hide' in flags:
            self.setVisible(False)
        # if 'dev' in flags:
        #     self.setVisible(False)
        if default is not None:
            self.value = default

    def setVisible(self, visible):
        if self.label is not None:
            self.label.setEnabled(visible)
            self.label.setVisible(visible)
        if self.editor is not None:
            self.editor.setEnabled(visible)
            self.editor.setVisible(visible)

    @property
    def value(self):
        return self._value

    def _find_option_idx(self, x):
        for idx, k in enumerate(self._options):
            if x in k:
                return idx
        raise RuntimeError(f'topic {self._topic}: option {x} not found')

    @value.setter
    def value(self, x):
        dtype = self._meta.get('dtype')
        b = self.editor.blockSignals(True)
        try:
            if 'ro' in self._meta.get('flags', []):
                if dtype == 'u32' and self._meta.get('format') == 'version':
                    x = int(x)
                    major = (x >> 24) & 0xff
                    minor = (x >> 16) & 0xff
                    patch = x & 0xffff
                    x = f'{major}.{minor}.{patch}'
                self.editor.setText(str(x))
            elif self._options is not None:
                idx = self._find_option_idx(x)
                self.editor.setCurrentIndex(idx)
            elif dtype == 'str':
                self.editor.setText(str(x))
            elif dtype in ['json', 'bin']:
                pass
            elif dtype in ['f32', 'f64']:
                pass
            elif dtype in ['u8', 'u16', 'u32', 'u64', 'i8', 'i16', 'i32', 'i64']:
                self.editor.setValue(int(x))
            elif dtype == 'bool':
                self.editor.setChecked(bool(x))
            else:
                pass
        finally:
            self._value = x
            self.editor.blockSignals(b)

    def _publish(self, value):
        self._parent.publish(self._topic, value)

    def _on_combobox(self, idx):
        opts = self._meta.get('options')[idx]
        self._publish(opts[0])

    def _on_clicked(self, value):
        self._publish(bool(value))

    def _on_spinbox(self, value):
        self._publish(value)


class PubSubWidget(QtWidgets.QWidget):

    def __init__(self, parent, device):
        self._device = weakref.ref(device)
        super(PubSubWidget, self).__init__(parent)
        self.setObjectName('pubsub_widget')
        self.setGeometry(QtCore.QRect(0, 0, 294, 401))

        self._rows = 0
        self._layout = QtWidgets.QGridLayout(self)
        self._layout.setSpacing(SPACING)
        self._layout.setContentsMargins(*MARGINS)
        self._layout.setObjectName('pubsub_widget_layout')
        self._values: dict[str, Value] = {}
        device.subscribe('', self._on_update, forward=True)

    def clear(self):
        while not self._layout.isEmpty():
            item = self._layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.setParent(None)
        self._values.clear()

    def publish(self, topic, value, retain=True):
        device = self._device()
        if device is None:
            return
        device.publish(topic, value, retain=retain, src_cbk=self._on_update)

    def get(self, topic):
        device = self._device()
        if device is None:
            return
        return device.get(topic)

    def _on_update(self, topic, value, retain=None):
        # print(f'{topic} : {value}')
        if topic == './conn/add':
            self._on_conn_add(value)
        elif topic == './conn/remove':
            self._on_conn_remove(value)
        elif topic.endswith('$') and topic != '$' and not topic.endswith('/$'):
            self._on_meta(topic[:-1], value)
        v = self._values.get(topic)
        if v is not None:
            v.value = value

    def _on_conn_add(self, topic_list):
        for topic in topic_list.split('\x1f'):
            if topic.endswith('/'):
                topic = topic[:-1]
            device = self._device()
            if device is not None:
                t = f'{topic}/$'
                log.info(f'request device metadata: {t}')
                device.publish(t, None)

    def _on_conn_remove(self, topic_list):
        for topic in topic_list.split('\x1f'):
            if topic.endswith('/'):
                topic = topic[:-1]
            topic = f'{topic}/'
            for value_str, value in self._values.items():
                if value_str.startswith(topic):
                    value.setVisible(False)

    def _on_meta(self, topic, meta):
        if meta is None:
            return
        if topic in self._values:
            self._values[topic].setVisible(True)
            return
        value = Value(self, topic, meta)
        if value.label is not None:
            self._values[topic] = value
            self._layout.addWidget(value.label, self._rows, 0)
            self._layout.addWidget(value.editor, self._rows, 1)
            self._rows += 1
            self._layout.update()
            try:
                value.value = self.get(topic)
            except Exception:
                pass  # use default value



# Copyright 2020-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pyfitterbap.comm.comm import Comm, LOG_TOPIC
from .resync import Resync
from .open import OpenDialog
from .device_widget import DeviceWidget
from .log_widget import LogWidget
from .waveform import WaveformWidget
from .vertical_scroll_area import VerticalScrollArea
from .expanding_widget import ExpandingWidget
from PySide6 import QtCore, QtGui, QtWidgets
from pyfitterbap import __version__, __url__
from pyfitterbap.pubsub import PubSub
import ctypes
import logging
import sys


log = logging.getLogger(__name__)
STATUS_BAR_TIMEOUT_DEFAULT = 2500
PORT_COUNT = 32


ABOUT = """\
<html>
<head>
</head>
<body>
FBP Comm UI<br/> 
Version {version}<br/>
<a href="{url}">{url}</a>

<pre>
Copyright 2018-2021 Jetperch LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
</pre>
</body>
</html>
"""


def menu_setup(d, parent=None):
    k = {}
    for name, value in d.items():
        name_safe = name.replace('&', '')
        if isinstance(value, dict):
            wroot = QtWidgets.QMenu(parent)
            wroot.setTitle(name)
            parent.addAction(wroot.menuAction())
            w = menu_setup(value, wroot)
            w['__root__'] = wroot
        else:
            w = QtGui.QAction(parent)
            w.setText(name)
            if callable(value):
                w.triggered.connect(value)
            parent.addAction(w)
        k[name_safe] = w
    return k


class Device(QtCore.QObject):

    def __init__(self, parent, pubsub, dev, baudrate=None):
        super(Device, self).__init__(parent)
        self._parent = parent
        self._dev = dev
        self._topic_prefix = dev + '/'
        self._pubsub = pubsub
        self._resync = Resync(self)
        self._on_device_publish = self._resync.wrap(self._subscribe_device)
        self.comm = None
        self.widget = None
        self._device_widget = None
        self.baudrate = baudrate

        if pubsub is not None:
            pubsub.subscribe(dev, self._subscribe_parent, forward=True)

    def __str__(self):
        return f'Device({self._dev})'

    @property
    def topic_prefix(self):
        return self._topic_prefix

    def _subscribe_parent(self, topic: str, value, retain=None, src_cbk=None):
        # forward from application pubsub to device's C pubsub
        log.info('ui → device: %s = %s', topic, value)
        if topic.startswith(self._topic_prefix):
            topic = topic[len(self._topic_prefix):]
            if self.comm is not None:
                self.comm.publish(topic, value, retain=retain, src_cbk=self._on_device_publish)

    def _subscribe_device(self, topic: str, value, retain=None, src_cbk=None):
        # forward from device's C pubsub to application pubsub
        topic = self._topic_prefix + topic
        if not topic.endswith('/din'):
            log.info("device → ui: %s = %s", topic, value)
        self._pubsub.publish(topic, value, retain=retain, src_cbk=self._subscribe_parent)

    def publish(self, topic: str, value, retain=None, src_cbk=None):
        topic = self._topic_prefix + topic
        self._pubsub.publish(topic, value, retain=retain, src_cbk=src_cbk)

    def get(self, topic: str):
        topic = self._topic_prefix + topic
        return self._pubsub.get(topic)

    def subscribe(self, topic, cbk, skip_retained=None, forward=None):
        def src_cbk_fn(topic: str, value, retain=None, src_cbk=None):
            if topic.startswith(self._topic_prefix):
                topic = topic[len(self._topic_prefix):]
                cbk(topic, value, retain=retain)
        self._pubsub.subscribe(self._topic_prefix + topic, src_cbk_fn, skip_retained=skip_retained, forward=forward)
        return src_cbk_fn

    def unsubscribe(self, topic, cbk):
        self._pubsub.unsubscribe(self._topic_prefix + topic, cbk)

    def open(self):
        self.close()
        try:
            self.comm = Comm(self._dev, self._on_device_publish, baudrate=self.baudrate)
            self.widget = ExpandingWidget(self._parent, str(self))
            self._device_widget = DeviceWidget(self.widget, self)
            self.widget.setWidget(self._device_widget)
        except Exception:
            log.exception('Could not open device')

    def close(self):
        if self.comm is not None:
            self.comm.close()
            self.comm = None
        if self._device_widget is not None:
            self._device_widget.close()
            self._device_widget = None
        self.widget = None

    def status_refresh(self):
        if self.comm is not None and self._device_widget is not None:
            status = self.comm.status()
            self._device_widget.status_update(status)


class DevicesWidget(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(DevicesWidget, self).__init__(parent)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                           QtWidgets.QSizePolicy.Expanding)
        self.setObjectName('devices_widget')
        self._index = 0
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setContentsMargins(5, 5, 5, 5)
        self._layout.setObjectName('devices_layout')

        self._spacer = QtWidgets.QSpacerItem(0, 0,
                                             QtWidgets.QSizePolicy.Minimum,
                                             QtWidgets.QSizePolicy.Expanding)
        self._layout.addItem(self._spacer)

    def add_device(self, device: Device):
        self._layout.insertWidget(self._index, device.widget)
        self._index += 1
        self._layout.update()

    def remove_device(self, device: Device):
        self._layout.removeWidget(device.widget)
        self._index -= 1


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, app):
        self._devices = {}
        self._pubsub = PubSub('ui')
        super(MainWindow, self).__init__()
        self.setObjectName('MainWindow')
        self.setWindowTitle('FBP Comm')
        self.resize(640, 480)

        self._central_widget = QtWidgets.QSplitter(QtCore.Qt.Vertical, self)
        self._central_widget.setObjectName('central')
        self._central_widget.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                           QtWidgets.QSizePolicy.Expanding)
        self.setCentralWidget(self._central_widget)

        self._hwidget = QtWidgets.QWidget(self._central_widget)
        self._central_widget.addWidget(self._hwidget)
        self._central_widget.setStretchFactor(0, 3)

        self._hlayout = QtWidgets.QHBoxLayout(self._hwidget)
        self._hlayout.setSpacing(6)
        self._hlayout.setContentsMargins(11, 11, 11, 11)
        self._hlayout.setObjectName('central_layout')

        self._devices_scroll = VerticalScrollArea(self._hwidget)
        self._hlayout.addWidget(self._devices_scroll)

        self._devices_widget = DevicesWidget(self._devices_scroll)
        self._devices_scroll.setWidget(self._devices_widget)

        self._waveform = WaveformWidget(self._hwidget, self._pubsub)
        size_policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        size_policy.setHorizontalStretch(1)
        self._waveform.setSizePolicy(size_policy)
        self._hlayout.addWidget(self._waveform)

        self._log_widget = LogWidget(self._central_widget)
        self._central_widget.addWidget(self._log_widget)
        self._central_widget.setStretchFactor(1, 1)

        # Status bar
        self._status_bar = QtWidgets.QStatusBar(self)
        self.setStatusBar(self._status_bar)
        self._status_indicator = QtWidgets.QLabel(self._status_bar)
        self._status_indicator.setObjectName('status_indicator')
        self._status_bar.addPermanentWidget(self._status_indicator)

        # status update timer
        self._status_update_timer = QtCore.QTimer(self)
        self._status_update_timer.setInterval(1000)  # milliseconds
        self._status_update_timer.timeout.connect(self._on_status_update_timer)
        self._status_update_timer.start()

        self._menu_bar = QtWidgets.QMenuBar(self)
        self._menu_items = menu_setup(
            {
                '&File': {
                    '&Open': self._on_file_open,
                    '&Close': self._on_file_close,
                    'E&xit': self._on_file_exit,
                },
                '&Help': {
                    '&Credits': self._on_help_credits,
                    '&About': self._on_help_about,
                }
            },
            self._menu_bar)
        self.setMenuBar(self._menu_bar)
        self.show()
        self._on_file_open()

    def _on_publish(self, topic, value, retain=None, src_cbk=None):
        log.info(f'publish {topic} => {value}')
        self._pubsub.publish(topic, value, retain=retain, src_cbk=src_cbk)

    def _device_open(self, dev, baudrate):
        self._device_close(dev)
        log.info('_device_open')
        try:
            device = Device(self, self._pubsub, dev, baudrate)
            device.subscribe(LOG_TOPIC, self._log_widget.on_publish, skip_retained=True)
            device.open()
            self._devices_widget.add_device(device)
            self._devices[dev] = device
        except Exception:
            log.exception('Could not open device')

    def _device_close(self, dev):
        log.info('_device_close')
        if dev in self._devices:
            device = self._devices.pop(dev)
            self._devices_widget.remove_device(device)
            try:
                device.unsubscribe(LOG_TOPIC, self._log_widget.on_publish)
                device.close()
            except Exception:
                log.exception('Could not close device')

    def _device_close_all(self):
        devs = list(self._devices.keys())
        for dev in devs:
            self._device_close(dev)

    def _on_file_open(self):
        log.info('_on_file_open')
        params = OpenDialog(self).exec_()
        if params is not None:
            log.info(f'open {params}')
            self._device_open(params['device'], params['baud'])

    def _on_status_update_timer(self):
        for device in self._devices.values():
            device.status_refresh()

    def closeEvent(self, event):
        log.info('closeEvent()')
        self._device_close_all()
        return super(MainWindow, self).closeEvent(event)

    def _on_file_close(self):
        log.info('_on_file_close')
        self._device_close_all()

    def _on_file_exit(self):
        log.info('_on_file_exit')
        self._device_close_all()
        self.close()

    def _on_help_credits(self):
        log.info('_on_help_credits')

    def _on_help_about(self):
        log.info('_on_help_about')
        txt = ABOUT.format(version=__version__,
                           url=__url__)
        QtWidgets.QMessageBox.about(self, 'Delta Link UI', txt)

    @QtCore.Slot(str)
    def status_msg(self, msg, timeout=None, level=None):
        """Display a status message.

        :param msg: The message to display.
        :param timeout: The optional timeout in milliseconds.  0
            does not time out.
        :param level: The logging level for the message.  None (default)
            is equivalent to log.INFO.
        """
        level = logging.INFO if level is None else level
        log.log(level, msg)
        timeout = STATUS_BAR_TIMEOUT_DEFAULT if timeout is None else int(timeout)
        self._status_bar.showMessage(msg, timeout)

    @QtCore.Slot(str)
    def error_msg(self, msg, timeout=None):
        self.status_msg(msg, timeout, level=log.ERROR)


def _high_dpi_enable():
    # http://doc.qt.io/qt-5/highdpi.html
    # https://vicrucann.github.io/tutorials/osg-qt-high-dpi/
    if sys.platform.startswith('win'):
        ctypes.windll.user32.SetProcessDPIAware()
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)


def run():
    _high_dpi_enable()
    app = QtWidgets.QApplication(sys.argv)
    ui = MainWindow(app)
    rc = app.exec_()
    del ui
    del app
    return rc



# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PySide6 import QtCore, QtGui, QtWidgets
from datetime import datetime

_LOG_LEVEL_MAP = {
    -1: ('O', 'off'),
    0:  ('E', 'emergency'),
    1:  ('A', 'alert'),
    2:  ('C', 'critical'),
    3:  ('E', 'error'),
    4:  ('W', 'warning'),
    5:  ('N', 'notice'),
    6:  ('I', 'info'),
    7:  ('D1', 'debug1'),
    8:  ('D2', 'debug2'),
    9:  ('D3', 'debug3'),
    10: ('A', 'all'),
}

_COLUMNS = ['Timestamp', 'Level', 'Device', 'Prefix', 'Thread', 'Filename', 'Line', 'Message']


class LogWidget(QtWidgets.QWidget):

    def __init__(self, parent=None):
        """Display Fitterbap device log messages.

        :param parent: The parent widget.
        """
        QtWidgets.QWidget.__init__(self, parent)
        self._model = QtGui.QStandardItemModel(0, 8, self)
        self._model.setHorizontalHeaderLabels(_COLUMNS)
        self._messages = []
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setSpacing(5)
        self._layout.setContentsMargins(5, 5, 5, 5)
        self._layout.setObjectName('log_layout')

        self._ctrl_bar = QtWidgets.QWidget(self)
        self._ctrl_bar.setObjectName('log_ctrl_bar')
        self._layout.addWidget(self._ctrl_bar)
        self._ctrl_layout = QtWidgets.QHBoxLayout(self._ctrl_bar)

        self._table = QtWidgets.QTableView(self)
        self._table.setObjectName('log_table')
        self._table.setModel(self._model)
        self._table.horizontalHeader().setStretchLastSection(True)
        self._table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self._layout.addWidget(self._table)
        self._scroll_timer_active = False

        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                           QtWidgets.QSizePolicy.Preferred)

    def _scroll_timer_start(self):
        if self._scroll_timer_active:
            return
        rows = self._model.rowCount()
        if rows <= 0:
            return
        last_row_pos = self._table.rowViewportPosition(rows - 1)
        height = self._table.viewport().height()
        if last_row_pos <= height:
            self._scroll_timer_active = True
            QtCore.QTimer.singleShot(10, self.scroll_to_bottom)

    @QtCore.Slot()
    def scroll_to_bottom(self):
        self._scroll_timer_active = False
        self._table.scrollToBottom()

    @QtCore.Slot(object)
    def message(self, msg):
        """Receive a new log message.

        :param msg: The log message dict which contains the fields:
            * timestamp: The float timestamp in UTC seconds from the POSIX epoch.
            * level: The log level integer.
            * device: The device string that generated this message.
            * origin_prefix: The prefix character for the subsystem that generated this message.
            * origin_thread: The thread integer for the subsystem that generated this message.
            * filename: The source code filename string.
            * line: The source code line number.
            * message: The emitted log message string.
        """
        utc = datetime.utcfromtimestamp(msg['timestamp'])
        utc_str = utc.isoformat()
        _, level_str = _LOG_LEVEL_MAP.get(msg['level'], _LOG_LEVEL_MAP[9])
        prefix = str(chr(msg['origin_prefix']))
        items = [
            QtGui.QStandardItem(utc_str),
            QtGui.QStandardItem(level_str),
            QtGui.QStandardItem(msg['device']),
            QtGui.QStandardItem(prefix),
            QtGui.QStandardItem(str(msg['origin_thread'])),
            QtGui.QStandardItem(msg['filename']),
            QtGui.QStandardItem(str(msg['line'])),
            QtGui.QStandardItem(msg['message']),
        ]
        msg['items'] = items
        self._messages.append(msg)
        self._model.appendRow(items)
        self._scroll_timer_start()

    def on_publish(self, topic, value, retain):
        self.message(value)
